from typing import Any
from unittest.mock import ANY

import pytest
from pikadantic.decorators import validate_body
from pikadantic.exceptions import PikadanticValidationError
from pydantic import BaseModel, RootModel


@pytest.fixture
def sample_model() -> type[BaseModel]:
    class SampleModel(BaseModel):
        text: str

    return SampleModel


@pytest.fixture
def sample_root_model() -> type[RootModel]:
    class SampleRootModel(RootModel):
        root: str

    return SampleRootModel


def callback(*args: Any) -> str:
    return "OK"


class TestValidateBody:
    def test_when_body_is_valid_then_callback_is_called(self, sample_model: type[BaseModel]):
        decorated = validate_body(sample_model)(callback)

        result = decorated(ANY, ANY, ANY, b'{"text": "test"}')

        assert result == "OK"

    def test_when_body_is_invalid_then_exception_is_raised(self, sample_model: type[BaseModel]):
        decorated = validate_body(sample_model)(callback)

        with pytest.raises(PikadanticValidationError):
            decorated(ANY, ANY, ANY, b'{"text": 1}')

    def test_when_body_is_invalid_and_raise_on_error_is_false_then_callback_is_not_called(
        self, sample_model: type[BaseModel]
    ):
        decorated = validate_body(sample_model, raise_on_error=False)(callback)

        result = decorated(ANY, ANY, ANY, b'{"text": 1}')

        assert result is None

    def test_when_body_is_valid_then_callback_is_called_with_root_model(self, sample_root_model: type[RootModel]):
        decorated = validate_body(sample_root_model, json=False)(callback)

        result = decorated(ANY, ANY, ANY, b"test")

        assert result == "OK"

    def test_when_root_model_and_json_is_true_then_exception_is_raised(self, sample_root_model: type[RootModel]):
        decorated = validate_body(sample_root_model, json=True)(callback)

        with pytest.raises(PikadanticValidationError):
            decorated(ANY, ANY, ANY, b'{"root": 1}')

    def test_when_root_model_and_json_is_true_and_raise_on_error_is_false_then_callback_is_not_called(
        self, sample_root_model: type[RootModel]
    ):
        decorated = validate_body(sample_root_model, json=True, raise_on_error=False)(callback)

        result = decorated(ANY, ANY, ANY, b'{"root": 1}')

        assert result is None



from functools import wraps
from typing import Callable, Optional, TypeVar

from pika.adapters.blocking_connection import BlockingChannel
from pika.spec import BasicProperties
from pydantic import BaseModel, ValidationError

from pikadantic.exceptions import PikadanticValidationError

R = TypeVar("R")

WrappedCallback = Callable[[BlockingChannel, str, BasicProperties, bytes], Optional[R]]


def validate_body(model: type[BaseModel], *, json: bool = True, raise_on_error: bool = True):
    """
    Decorator that validates a message body against a Pydantic model before calling the handler.

    Args:
        model (Type[BaseModel]): Pydantic model class used for validation.
        json (bool): If True, parses body as JSON. If False, uses raw data. Default is True.
        raise_on_error (bool): If True, raises PikadanticValidationError on failure. Otherwise, skips handler.
        Default is True.

    Returns:
        Callable: The decorated handler function.

    Raises:
        PikadanticValidationError: If validation fails and raise_on_error is True.
    """

    def decorator(func: WrappedCallback[R]) -> WrappedCallback[Optional[R]]:
        @wraps(func)
        def wrapper(
            channel: BlockingChannel,
            method: str,
            properties: BasicProperties,
            body: bytes,
        ) -> Optional[R]:
            try:
                if json:
                    model.model_validate_json(body)
                else:
                    model.model_validate(body)
            except ValidationError as e:
                if raise_on_error:
                    raise PikadanticValidationError(e) from e
                return None

            return func(channel, method, properties, body)

        return wrapper

    return decorator


__all__ = ["validate_body"]



from pydantic import ValidationError


class PikadanticBaseError(Exception):
    pass


class PikadanticValidationError(PikadanticBaseError):
    def __init__(self, error: ValidationError) -> None:
        self.error = error

    def __str__(self) -> str:
        return f"Validation error: {self.error}"

    def __repr__(self) -> str:
        return f"PikadanticValidationError(error={self.error!r})"
